from __future__ import annotations

import asyncio
import logging
import time
from typing import Any, Optional, Dict, List, Union

from pyobs.object import Object
from .pilarerror import PilarError


log = logging.getLogger(__name__)


class PilarCommand(object):
    def __init__(self, command: str):
        self.command = command
        self.id: Optional[int] = None
        self.time: Optional[float] = None
        self.sent = False
        self.acknowledged = False
        self.completed = asyncio.Event()
        self.error: Optional[str] = None
        self.data = Optional[Any]
        self.values: Dict[str, Any] = {}

    def __call__(self, transport: asyncio.Transport) -> None:
        # send command
        cmd = str(self.id) + ' ' + self.command + '\n'
        transport.write(bytes(cmd, 'utf-8'))

        # store current time and set as sent
        self.time = time.time()
        self.sent = True

    def parse(self, line: str) -> None:
        # split line and check ID
        s = line.split()
        if int(s[0]) != self.id:
            return

        # acknowledge
        if 'COMMAND OK' in line or 'COMMAND ERROR' in line:
            self.acknowledged = True
            if 'COMMAND ERROR' in line:
                self.error = line

        # payload
        if 'DATA INLINE' in line:
            # get key and value
            pos = line.find('=')
            key = line[line.find('DATA INLINE') + 12:pos]
            value = line[pos+1:]
            # type?
            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
            # store it
            self.values[key] = value
            # we always store the last result as data, makes it easier for commands requesting only a single value
            self.data = value

        # finish
        elif 'COMMAND COMPLETE' in line or 'COMMAND FAILED' in line:
            self.completed.set()

    async def wait(self, timeout: int = 5) -> None:
        """Wait for the command to finish.

        Args:
            timeout: Timeout for waiting in seconds.
        """
        await asyncio.wait_for(self.completed.wait(), timeout)


class PilarClientProtocol(asyncio.Protocol):
    """ asyncio.Protocol implementation for the Pilar interface. """

    def __init__(self, driver: PilarDriver, loop: asyncio.AbstractEventLoop, username: str, password: str):
        """ Creates a SicamTcpClientProtocol.
        :param driver: SicamTcpDriver instance.
        :param loop: asyncio event loop.
        :param username: Username for login.
        :param password: Password for login
        :return:
        """

        # init some stuff
        self._driver = driver
        self._buffer = ''
        self._loop = loop
        self._transport: Optional[asyncio.Transport] = None
        self._username = username
        self._password = password
        self._logged_in = False
        self._id: int = 0
        self._commands: List[PilarCommand] = []

        # store self in driver
        self._driver.protocol = self

    @property
    def logged_in(self) -> float:
        return self._logged_in

    def connection_made(self, transport):
        """ Called, when the protocol is connected to the server.
        :param transport: Transport connected to server.
        :return:
        """

        # store transport
        self._transport = transport

    async def stop(self) -> None:
        """ Disconnect gracefully. """

        if self._transport:
            # send disconnect
            log.info('Sending disconnect...')
            self._transport.write(b'disconnect')

            # disconnect
            self._transport.close()
            log.info('Disconnected from pilar.')

    def data_received(self, data: bytes) -> None:
        """ Called, when new data arrives from the server.
        :param data: New chunk of data.
        :return:
        """

        # this cannot happen, but let's make MyPy happy
        if self._transport is None:
            return

        # add data to buffer
        self._buffer += data.decode('utf-8')

        # create as many packets as possible
        while self._buffer and '\n' in self._buffer:
            # extract line from buffer
            length = self._buffer.find('\n')
            line = self._buffer[:length]
            self._buffer = self._buffer[length+1:]

            # AUTH?
            if 'AUTH PLAIN' in line:
                log.info('Logging into Pilar...')
                # send AUTH line
                auth = 'AUTH PLAIN "' + self._username + '" "' \
                       + self._password + '"\n'
                self._transport.write(bytes(auth, 'utf-8'))

            elif 'AUTH OK' in line:
                log.info('Authentication for Pilar successful.')
                self._logged_in = True

            elif 'AUTH FAILED' in line:
                log.warning('Authentication for Pilar failed.')
                self._logged_in = False

            else:
                # loop all commands and parse line
                commands_to_delete = []
                for cmd in self._commands:
                    # parse line
                    cmd.parse(line)

                    # is command finished?
                    if cmd.completed.is_set():
                        # remove it from list
                        commands_to_delete.append(cmd)

                # delete finished commands
                for cmd in commands_to_delete:
                    self._commands.remove(cmd)

    def execute(self, command: str) -> PilarCommand:
        if self._transport is None:
            raise RuntimeError()

        # get next id
        self._id += 1

        # create command
        cmd = PilarCommand(command)
        cmd.id = self._id
        self._commands.append(cmd)

        # execute return
        cmd(self._transport)
        return cmd


class PilarDriver(Object):
    """ Wrapper for easy communication with Pilar. """

    def __init__(self, host: str, port: int, username: str, password: str, **kwargs: Any):
        """ Create new driver. """
        Object.__init__(self, **kwargs)

        # init some stuff
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._filters: List[str] = []
        self.protocol: Optional[PilarClientProtocol] = None

        # errors
        self._has_error = False
        self._error_thread = None

        # background tasks
        self.add_background_task(self._error_background_task)

    async def open(self) -> None:
        """ Open connection to SIImage. """

        # create connection
        loop = asyncio.get_running_loop()
        await loop.create_connection(lambda: PilarClientProtocol(self, loop, self._username, self._password),
                                     self._host, self._port)

        # wait for login
        while self.protocol is None or not self.protocol.logged_in:
            await asyncio.sleep(0.1)

    async def close(self) -> None:
        """ Close connection to SIImage. """

        # safely close the connection
        if self.protocol:
            await self.protocol.stop()

    async def _error_background_task(self) -> None:
        # run until closing
        while True:
            # not logged in?
            if self.protocol is None or not self.protocol.logged_in:
                await asyncio.sleep(5)

            # check for errors and clear them
            self._has_error = not await self.clear_errors()

            # on error, wait and check
            if self._has_error:
                # wait a little
                await asyncio.sleep(5)

                # check again
                self._has_error = not await self.check_errors()

            # wait five seconds
            await asyncio.sleep(5)

    @property
    def has_error(self) -> bool:
        return self._has_error

    @property
    def is_open(self) -> bool:
        """ Whether connection is open."""
        return self.protocol is not None

    async def get(self, key: str) -> Any:
        if self.protocol is None:
            raise RuntimeError()
        cmd = self.protocol.execute('GET ' + key)
        await cmd.wait()
        return cmd.data

    async def get_multi(self, keys: List[str]) -> Dict[str, Any]:
        if self.protocol is None:
            raise RuntimeError()
        # join keys with ";" and execute
        cmd = self.protocol.execute('GET ' + ';'.join(keys))
        await cmd.wait()
        return cmd.values

    async def set(self, key: str, value: Any, wait: bool = True, timeout: int = 5000) -> Union[bool, PilarCommand]:
        """Set a variable with a given value.

        Args:
            key: Name of variable to set.
            value: New value.
            wait: Whether or not to wait for command.
            timeout: Timeout for waiting.
        """
        if self.protocol is None:
            raise RuntimeError()

        # execute SET command
        cmd = self.protocol.execute(f'SET {key}="{str(value)}"')

        # want to wait?
        if wait:
            await cmd.wait(timeout=timeout)
            return cmd.error is None

        # return cmd
        return cmd

    async def safe_set(self, key: str, value: Any, timeout: int = 5000, msg: str = '') -> None:
        """Set a variable with a given value, raise exception on error.

        Args:
            key: Name of variable to set.
            value: New value.
            timeout: Timeout for waiting.
            msg: Message to add to exception text.
        """
        if self.protocol is None:
            raise RuntimeError()

        # execute SET command
        cmd = self.protocol.execute(f'SET {key}="{str(value)}"')

        # wait
        await cmd.wait(timeout=timeout)
        if cmd.error is not None:
            raise ValueError(msg + cmd.error)

    async def list_errors(self) -> List[PilarError]:
        """Fetch list of errors from telescope.

        From OpenTSI documentation about TELESCOPE.STATUS.LIST:
            A comma separated list of function groups that currently have
            problems in the following format:
            <group>|<level>[:<component>|<level>[;<component>...]]
            [:<error>|<detail>|<level>|<component>[;<error>...]]
            [,<group>...]
            <group> One of the above listed function groups
            <level> Bitwise “OR” of all errors in the group resp. compo-
            nent or for the individual error. The bits have the same
            meaning as for GLOBAL.
            <component> The OpenTCI module name (possibly includ-
            ing a index in []).
            <error> The hardware specific error code
            <detail> The hardware specific detail information for the
            error code.
            The information from <error>/<detail> should only be used
            for logging, as it is hardware specific and may change at any
            time.
            At most one entry per group is generated. If the delimiters
            should occur within the names or messages, they will be either
            escaped with a backslash or the entire entry is put in double
            quotes.
        """

        # init error list
        error_list: List[PilarError] = []

        # get list of errors
        errors = await self.get('TELESCOPE.STATUS.LIST')

        # divide into groups and loop them
        for group in errors.split(','):
            # find last colon and split everything after that at semicolon
            errors = group[group.rfind(':')+1:].split(';')

            # loop errors
            for error in errors:
                # split by | to get name of error
                name = error.split('|')[0]
                if len(name) == 0:
                    continue

                # create error and add to list
                err = PilarError.create(name)
                if err is None:
                    raise ValueError('Error while handling error.')
                else:
                    error_list.append(err)

        # return list of errors
        return error_list

    async def clear_errors(self) -> bool:
        """Clears Pilar errors."""

        # get telescope status
        level = int(await self.get('TELESCOPE.STATUS.GLOBAL'))

        # check level
        if level & (0x01 | 0x02):
            log.warning('Found severe errors with level %d.', level)
        else:
            return True
            #log.info('Current error level is %d.', level)

        # check fatal errors
        await self.list_errors()
        if PilarError.check_fatal():
            log.error('Cannot clear errors, fatal condition.')
            return False

        # do clearing
        log.info('Clearing telescope errors...')
        await self.set('TELESCOPE.STATUS.CLEAR', level)
        return True

    async def check_errors(self) -> bool:
        """Check for errors after clearing."""

        # get telescope status
        level = int(await self.get('TELESCOPE.STATUS.GLOBAL'))

        # check level
        if level & (0x01 | 0x02):
            log.error('Could not clear severe errors.')
            return False
        else:
            log.info('Remaining error level is %d.', level)
            return True

    async def init(self, attempts: int = 3, wait: float = 10., attempt_timeout: float = 30.) -> bool:
        """Initialize telescope

        Args:
            attempts (int): Number of attempts for initializing telescope.
            wait (float):   Wait time in seconds after sending command.
            attempt_timeout (float): Number of seconds to allow for each attempt.

        Returns:

        """

        # check, whether telescope is initialized already
        ready = await self.get('TELESCOPE.READY_STATE')
        if float(ready) == 1.:
            log.info("Telescope already initialized.")
            return True

        # we give it a couple of attempts
        log.info("Initializing telescope...")
        for attempt in range(attempts):
            # init telescope
            await self.set('TELESCOPE.READY', 1)

            # sleep a little
            await asyncio.sleep(wait)

            # wait for init
            waited = 0.
            while waited < attempt_timeout:
                # get status
                ready = await self.get('TELESCOPE.READY_STATE')
                if float(ready) == 1.:
                    log.info("Telescope initialized.")
                    return True

                # sleep  a little
                waited += 0.5
                await asyncio.sleep(0.5)

        # we should never arrive here
        log.error('Could not initialize telescope.')
        return False

    async def park(self, attempts: int = 3, wait: float = 10., attempt_timeout: float = 30.) -> bool:
        # check, whether telescope is parked already
        ready = await self.get('TELESCOPE.READY_STATE')
        if float(ready) == 0.:
            log.info("Telescope already parked.")
            return True

        # we give it a couple of attempts
        log.info("Parking telescope...")
        for attempt in range(attempts):
            # parking telescope
            await self.set('TELESCOPE.READY', 0)

            # sleep a little
            await asyncio.sleep(wait)

            # wait for init
            waited = 0.
            while waited < attempt_timeout:
                # get status
                ready = await self.get('TELESCOPE.READY_STATE')
                if float(ready) == 0.:
                    log.info("Telescope parked.")
                    return True

        # we should never arrive here
        log.error('Could not park telescope.')
        return False

    async def reset_focus_offset(self) -> None:
        # get focus and offset
        focus = float(await self.get('POSITION.INSTRUMENTAL.FOCUS.TARGETPOS'))
        offset = float(await self.get('POSITION.INSTRUMENTAL.FOCUS.OFFSET'))

        # need to do something?
        if abs(offset) > 1e-5:
            # set new
            cmd1 = await self.set('POSITION.INSTRUMENTAL.FOCUS.TARGETPOS', focus + offset, wait=False)
            cmd2 = await self.set('POSITION.INSTRUMENTAL.FOCUS.OFFSET', 0, wait=False)

    async def focus(self, position: float, timeout: int = 30000, accuracy: float = 0.01, sleep: int = 500,
                    retry: int = 3, sync_thermal: bool = False, sync_port: bool = False, sync_filter: bool = False,
                    disable_tracking: bool = False, abort_event: Optional[asyncio.Event] = None) -> bool:
        # reset any offset
        #self.reset_focus_offset()

        # set sync_mode, first bit is always set
        sync_mode = 1
        if sync_thermal:
            log.info('Enabling synchronization with thermal model...')
            sync_mode |= 1 << 1
        if sync_port:
            log.info('Enabling synchronization with port specific offset...')
            sync_mode |= 1 << 2
        if sync_filter:
            log.info('Enabling synchronization with filter specific offset...')
            sync_mode |= 1 << 3
        if disable_tracking:
            log.info('Turning off focus motor during tracking...')
            sync_mode |= 1 << 4

        # setting new focus
        log.info('Setting new focus value to %.3fmm...', position)
        await self.set('POINTING.SETUP.FOCUS.SYNCMODE', sync_mode)
        #self.set('POSITION.INSTRUMENTAL.FOCUS.OFFSET', 0)
        await self.set('POINTING.SETUP.FOCUS.POSITION', position)
        await self.set('POINTING.TRACK', 4)

        # loop until finished
        delta = 1e10
        waited = 0
        attempts = 0
        while delta >= accuracy:
            # abort?
            if abort_event is not None and abort_event.is_set():
                return False

            # sleep a little
            await asyncio.sleep(sleep / 1000.)
            waited += sleep

            # get focus distance
            delta = abs(float(await self.get('POSITION.INSTRUMENTAL.FOCUS.TARGETDISTANCE')))
            log.info('Distance to new focus: %.3fmm.', delta)

            # waiting too long?
            if waited > timeout:
                # got more retries?
                if attempts < retry:
                    # yes, so try again
                    attempts += 1
                    waited = 0
                    log.warning('Focus timeout, starting attempt %d.', attempts+1)
                    await self.set('POINTING.SETUP.FOCUS.POSITION', position)
                    await self.set('POINTING.TRACK', 4)

                else:
                    # no, we're out of time
                    log.error('Focusing not possible.')
                    return False

        # get new focus
        foc = await self.get('POSITION.INSTRUMENTAL.FOCUS.REALPOS')
        log.info('New focus position reached: %.3fmm.', float(foc))
        return True

    async def goto(self, alt: float, az: float, abort_event: asyncio.Event) -> bool:
        # stop telescope
        await self.set('POINTING.TRACK', 0)

        # prepare derotator
        await self.set('POINTING.SETUP.DEROTATOR.SYNCMODE', 3)
        await self.set('POINTING.SETUP.DEROTATOR.OFFSET', 0.)

        # set new coordinates
        await self.set('OBJECT.HORIZONTAL.AZ', az)
        await self.set('OBJECT.HORIZONTAL.ALT', alt)

        # start moving
        await self.set('POINTING.TRACK', 2)

        # wait for it
        success = False
        for attempt in range(5):
            # abort?
            if abort_event.is_set():
                return False

            # wait
            success = await self._wait_for_value('TELESCOPE.MOTION_STATE', '0', abort_event=abort_event)
            if success:
                break

            # sleep a little and try again
            await asyncio.sleep(1)
            await self.set('POINTING.TRACK', 2)
            log.warning('Attempt %d for moving to position failed.', attempt + 1)

        # success
        return success

    async def track(self, ra: float, dec: float, abort_event: asyncio.Event) -> bool:
        # stop telescope
        await self.set('POINTING.TRACK', 0)

        # prepare derotator
        await self.set('POINTING.SETUP.DEROTATOR.SYNCMODE', 3)
        await self.set('POINTING.SETUP.DEROTATOR.OFFSET', 0.)

        # set new coordinates
        await self.set('OBJECT.EQUATORIAL.RA', ra/15.)
        await self.set('OBJECT.EQUATORIAL.DEC', dec)

        # start tracking
        await self.set('POINTING.TRACK', 1)

        # wait for it
        success = False
        for attempt in range(5):
            # abort?
            if abort_event.is_set():
                return False

            # wait
            success = await self._wait_for_value('TELESCOPE.MOTION_STATE', '11', '0', abort_event=abort_event)
            if success:
                break

            # got any errors?
            if len(await self.list_errors()) > 0:
                return False

            # sleep a little and try again
            await asyncio.sleep(1)
            await self.set('POINTING.TRACK', 2)
            log.warning('Attempt %d for moving to position failed.', attempt + 1)

        # success
        return success

    async def _wait_for_value(self, var: str, value: Any, not_value: Optional[Any] = None,
                              abort_event: Optional[asyncio.Event] = None) -> bool:
        # sleep a little
        await asyncio.sleep(0.5)

        while True:
            # abort?
            if abort_event is not None and abort_event.is_set():
                return False

            # get variable
            val = float(await self.get(var))

            # check
            if val == float(value):
                return True
            elif not_value is not None and val == float(not_value):
                return False

            # sleep a little
            await asyncio.sleep(1)

    async def fits_data(self) -> Dict[str, float]:
        return {
            'TEL-T1': float(await self.get('AUXILIARY.SENSOR[3].VALUE')),
            'TEL-T2': float(await self.get('AUXILIARY.SENSOR[1].VALUE')),
            'TEL-T3': float(await self.get('AUXILIARY.SENSOR[2].VALUE')),
            'TEL-T4': float(await self.get('AUXILIARY.SENSOR[4].VALUE')),
            'TEL-FOCU': float(await self.get('POSITION.INSTRUMENTAL.FOCUS.REALPOS'))
        }

    async def init_filters(self) -> None:
        # get number of filters
        num = int(await self.get('TELESCOPE.CONFIG.PORT[2].FILTER'))
        log.info("Found %d filters.", num)

        # loop all filters
        self._filters = []
        for i in range(num):
            # set filter
            await self.set('POINTING.SETUP.FILTER.INDEX', i)

            # get filter name
            name = await self.get('POINTING.SETUP.FILTER.NAME')

            # strip quotes
            name = name.replace('"', '')

            # append to list
            log.info('Found filter %s.', name)
            self._filters.append(name)

    async def filters(self) -> List[str]:
        if not self._filters:
            await self.init_filters()
        return self._filters

    async def change_filter(self, filter_name: str, force_forward: bool = True,
                            abort_event: Optional[asyncio.Event] = None) -> bool:
        # get current filter id
        cur_id = int(float(await self.get('POSITION.INSTRUMENTAL.FILTER[2].CURRPOS')))

        # find ID of filter
        filter_id = self._filters.index(filter_name)
        if filter_id == cur_id:
            return True
        log.info('Changing to filter %s with ID %d.', filter_name, filter_id)

        # force only forward motion? if new ID is smaller than current one, first move to last filter
        if force_forward:
            # do until we're at the current filter
            while cur_id != filter_id:
                # how far can we go?
                for i in range(3):
                    # increase cur filter by one, wrap at end
                    cur_id += 1
                    if cur_id >= len(self._filters):
                        cur_id = 0

                    # got it?
                    if cur_id == filter_id:
                        break

                # move there
                if not self._change_filter_to_id(cur_id, abort_event):
                    log.info('Could not change to filter.')
                    return False

            # finished
            log.info('Successfully changed to filter %s.', filter_name)
            return True

        else:
            # simply go to requested filter
            if self._change_filter_to_id(filter_id, abort_event):
                log.info('Successfully changed to filter %s.', self._filters[filter_id])
                return True
            else:
                log.info('Could not change filter.')
                return False

    async def _change_filter_to_id(self, filter_id: int, abort_event: Optional[asyncio.Event]= None) -> bool:
        # set it
        await self.set('POINTING.SETUP.FILTER.INDEX', filter_id)
        await self.set('POINTING.TRACK', 3)

        # wait for it
        return await self._wait_for_value('POSITION.INSTRUMENTAL.FILTER[2].CURRPOS',
                                          filter_id, abort_event=abort_event)

    async def filter_name(self, filter_id: Optional[int] = None) -> str:
        if filter_id is None:
            filter_id = int(float(await self.get('POSITION.INSTRUMENTAL.FILTER[2].CURRPOS')))
        return self._filters[filter_id]

    async def stop(self) -> None:
        # stop telescope
        # TODO: there is obviously some kind of ABORT command, look into it

        # deactivate tracking
        await self.set('POINTING.TRACK', 0)



from __future__ import annotations

import datetime
import logging
from typing import Dict, NamedTuple, Optional, List

log = logging.getLogger(__name__)


class ErrorBehaviour(NamedTuple):
    ignore: bool = False
    reset_max: int = 10
    reset_timeout: int = 300
    accum_max: int = 5
    accum_span: int = 86400


# List of errors
ERRORS = {
    'ERR_GPS_PositionLost': ErrorBehaviour(ignore=True),
    'ERR_GPS_LeapSecond': ErrorBehaviour(ignore=True),
    'ERR_GPS_TooFewSatellites': ErrorBehaviour(ignore=True),
    'ERR_FilterWheel_RefMismatch': ErrorBehaviour(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600),
    'ERR_Elevation_ETELExecError': ErrorBehaviour(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600),
    'ERR_Oil_TemperatureLow': ErrorBehaviour(ignore=True),
    'ERR_Oil_NoExtractionTimeout': ErrorBehaviour(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600),
    'ERR_Brake_ClosedFromOther': ErrorBehaviour(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600),
    'ERR_Azimuth_ETELError': ErrorBehaviour(ignore=False, reset_timeout=0, accum_max=5, accum_span=3600)
}


class PilarError(object):
    _errors: Dict[str, PilarError] = {}

    def __init__(self, name: str):
        """Initializes an error

        Args:
            name:           Name of error.
        """
        self._name = name
        self._behaviour = ERRORS[name]
        self._dates: List[datetime.datetime] = []

    @property
    def name(self) -> str:
        return self._name

    @property
    def ignore(self) -> bool:
        return self._behaviour.ignore

    @property
    def reset_max(self) -> int:
        return self._behaviour.reset_max

    @property
    def reset_timeout(self) -> float:
        return self._behaviour.reset_timeout

    @property
    def accum_max(self) -> int:
        return self._behaviour.accum_max

    @property
    def accum_span(self) -> float:
        return self._behaviour.accum_span

    def occur(self) -> bool:
        """Should be called whenever error occurs.

        Returns:
            (bool) Whether or not it is safe to continue.
        """

        # ignore it?
        if self.ignore:
            return True

        # add now to list of dates
        self._dates.append(datetime.datetime.utcnow())
        return True

    def fatal(self) -> bool:
        # maximum number reached?
        if len(self._dates) > self.reset_max:
            log.warning('Error %s occurred more often than allowed (%d times).', self._name, len(self._dates))
            return True

        # time between last two errors?
        if len(self._dates) >= 2:
            # calculate difference between last two errors
            diff = (self._dates[-1] - self._dates[-2]).total_seconds()
            # check
            if diff < self.reset_timeout:
                log.warning('Duration (%.0fs) between last two errors of %s was shorter than allowed.',
                            diff, self._name)
                return True

        # collect errors within the last accum_span
        accum_errors = [d for d in self._dates if (self._dates[-1] - d).total_seconds() < self.accum_span]
        if len(accum_errors) > self.accum_max:
            log.warning('Too many (%d) errors of %s occurred during last %d seconds.',
                        len(accum_errors), self._name, self.accum_span)
            return True

        # everything okay
        return False

    @staticmethod
    def create(error_name: str) -> Optional[PilarError]:
        # error not found?
        if error_name not in ERRORS:
            log.error('Unknown error "%s" occurred.', error_name)
            return None

        # TODO: add more info
        log.warning('Error "%s" occurred.', error_name)

        # did we encounter this error before?
        if error_name not in PilarError._errors:
            # create new one and add to list
            err = PilarError(error_name)
            PilarError._errors[error_name] = err

        # get error and let error occur
        err = PilarError._errors[error_name]
        err.occur()

        # return error
        return err

    @staticmethod
    def check_fatal() -> bool:
        # loop all errors
        for err in PilarError._errors.values():
            if err.fatal():
                return True

        # no fatal condition found
        return False



import asyncio
import datetime
import logging
import os.path
from typing import Tuple, List, Dict, Any, Optional

from pyobs.mixins import FitsNamespaceMixin
from pyobs.events import FilterChangedEvent, OffsetsAltAzEvent
from pyobs.interfaces import IFilters, IFocuser, ITemperatures, IOffsetsAltAz, IPointingSeries
from pyobs.modules import timeout
from pyobs.modules.telescope.basetelescope import BaseTelescope
from pyobs.utils.enums import MotionStatus
from pyobs.utils.threads import LockWithAbort
from .pilardriver import PilarDriver

log = logging.getLogger(__name__)


# TODO: use asyncio in driver directly
class PilarTelescope(BaseTelescope, IOffsetsAltAz, IFilters, IFocuser, ITemperatures, IPointingSeries,
                     FitsNamespaceMixin):
    def __init__(self, host: str, port: int, username: str, password: str,
                 pilar_fits_headers: Optional[Dict[str, Any]] = None, temperatures: Optional[Dict[str, str]] = None,
                 force_filter_forward: bool = True, pointing_path: Optional[str] = None, **kwargs: Any):
        BaseTelescope.__init__(self, **kwargs, motion_status_interfaces=['ITelescope', 'IFilters', 'IFocuser'])

        # add thread func
        self.add_background_task(self._pilar_update)

        # init
        self._pilar_connect = host, port, username, password
        self._filters: List[str] = []
        self._force_filter_forward = force_filter_forward
        self._pilar_fits_headers = pilar_fits_headers if pilar_fits_headers else {}
        self._temperatures = temperatures if temperatures else {}
        self._temperatures = {}

        # pilar
        self._pilar_connect = host, port, username, password
        log.info('Connecting to Pilar at %s:%d...', host, port)
        self._pilar = self.add_child_object(PilarDriver, host=host, port=port, username=username, password=password)

        # create update thread
        self._status: Dict[str, Any] = {}

        # optimal focus
        self._last_focus_time = None

        # some multi-threading stuff
        self._lock_focus = asyncio.Lock()
        self._abort_focus = asyncio.Event()
        self._lock_filter = asyncio.Lock()
        self._abort_filter = asyncio.Event()

        # pointing
        self._pointing_path = pointing_path
        self._pointing_id = 1

        # set Pilar variables for status updates...
        self._pilar_variables = [
            'OBJECT.EQUATORIAL.RA', 'OBJECT.EQUATORIAL.DEC',
            'POSITION.EQUATORIAL.RA_J2000', 'POSITION.EQUATORIAL.DEC_J2000',
            'POSITION.HORIZONTAL.ZD', 'POSITION.HORIZONTAL.ALT', 'POSITION.HORIZONTAL.AZ',
            'POSITION.INSTRUMENTAL.FOCUS.REALPOS',
            'POSITION.INSTRUMENTAL.FILTER[2].CURRPOS',
            'POSITION.INSTRUMENTAL.DEROTATOR[2].REALPOS', 'POINTING.SETUP.DEROTATOR.OFFSET',
            'TELESCOPE.READY_STATE', 'TELESCOPE.MOTION_STATE',
            'POSITION.INSTRUMENTAL.AZ.OFFSET', 'POSITION.INSTRUMENTAL.ZD.OFFSET'
        ]

        # ... and add user defined ones
        for var in self._pilar_fits_headers.keys():
            if var not in self._pilar_variables:
                self._pilar_variables.append(var)

        # ... and temperatures
        for var in self._temperatures.values():
            if var not in self._pilar_variables:
                self._pilar_variables.append(var)

        # mixins
        FitsNamespaceMixin.__init__(self, **kwargs)

    async def open(self) -> None:
        """Open module."""
        await BaseTelescope.open(self)

        # get list of filters
        self._filters = await self._pilar.filters()

        # subscribe to events
        if self.comm:
            await self.comm.register_event(FilterChangedEvent)

    async def close(self) -> None:
        await BaseTelescope.close(self)

        if self._pilar is not None:
            log.info('Closing connection to Pilar...')
            await self._pilar.close()
        log.info('Shutting down...')

    async def _pilar_update(self) -> None:
        # log
        log.info('Starting Pilar update thread...')

        while not self.closing.is_set():
            # no pilar connection yet?
            if self._pilar is None or not self._pilar.is_open:
                await asyncio.sleep(1)
                continue

            # catch everything
            try:
                # do nothing on error
                if self._pilar.has_error:
                    await asyncio.sleep(10)
                    continue

                # define values to request
                keys = self._pilar_variables

                # get data
                try:
                    multi = await self._pilar.get_multi(keys)
                except TimeoutError:
                    # sleep a little and continue
                    log.error('Request to Pilar timed out.')
                    await asyncio.sleep(60)
                    continue

                # check for ready state
                if 'TELESCOPE.READY_STATE' not in multi:
                    await asyncio.sleep(1)
                    continue

                # set status
                self._status = {}
                for key in keys:
                    try:
                        self._status[key] = float(multi[key])
                    except ValueError:
                        log.warning(f'Could not find {key} in response from Pilas.')

                # set motion status
                # we always set PARKED, INITIALIZING, ERROR, the others only on init
                if float(self._status['TELESCOPE.READY_STATE']) == 0.:
                    await self._change_motion_status(MotionStatus.PARKED)
                elif 0. < float(self._status['TELESCOPE.READY_STATE']) < 1.:
                    await self._change_motion_status(MotionStatus.INITIALIZING)
                elif float(self._status['TELESCOPE.READY_STATE']) < 0.:
                    await self._change_motion_status(MotionStatus.ERROR)
                else:
                    # telescope is initialized, check motion state
                    ms = int(self._status['TELESCOPE.MOTION_STATE'])
                    if ms & (1 << 0):
                        # first bit indicates moving
                        if await self.get_motion_status() == MotionStatus.UNKNOWN:
                            await self._change_motion_status(MotionStatus.SLEWING)
                    elif ms & (1 << 2):
                        # third bit indicates tracking
                        if await self.get_motion_status() == MotionStatus.UNKNOWN:
                            await self._change_motion_status(MotionStatus.TRACKING)
                    else:
                        # otherwise we're idle
                        if await self.get_motion_status() == MotionStatus.UNKNOWN:
                            await self._change_motion_status(MotionStatus.IDLE)

                # sleep a second
                await asyncio.sleep(1)

            except:
                log.exception('An unexpected error occured.')
                await asyncio.sleep(10)

        # log
        log.info('Shutting down Pilar update thread...')

    async def get_fits_header_before(self, namespaces: Optional[List[str]] = None, **kwargs: Any) \
            -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # get headers from base
        hdr = await BaseTelescope.get_fits_header_before(self)

        # define values to request
        keys = {
            'TEL-FOCU': ('POSITION.INSTRUMENTAL.FOCUS.REALPOS', 'Focus position [mm]'),
            'TEL-ROT': ('POSITION.INSTRUMENTAL.DEROTATOR[2].REALPOS', 'Derotator instrumental position at end [deg]'),
            'DEROTOFF': ('POINTING.SETUP.DEROTATOR.OFFSET', 'Derotator offset [deg]'),
            'AZOFF': ('POSITION.INSTRUMENTAL.AZ.OFFSET', 'Azimuth offset'),
            'ALTOFF': ('POSITION.INSTRUMENTAL.ZD.OFFSET', 'Altitude offset')
        }

        # add ones from config
        for var, h in self._pilar_fits_headers.items():
            keys[h[0]] = (var, h[1])

        # Monet/S: 3=T1, 1=T2
        # Monet/N: 2=T1, 1=T2

        # create dict and add alt and filter
        status = self._status.copy()
        for key, v in keys.items():
            if v[0] in status:
                hdr[key] = (status[v[0]], v[1])

        # negative ALTOFF
        hdr['ALTOFF'] = (-hdr['ALTOFF'][0], hdr['ALTOFF'][1])

        # filter
        if 'POSITION.INSTRUMENTAL.FILTER[2].CURRPOS' in status:
            filter_id = status['POSITION.INSTRUMENTAL.FILTER[2].CURRPOS']
            hdr['FILTER'] = (await self._pilar.filter_name(filter_id), 'Current filter')

        # return it
        return self._filter_fits_namespace(hdr, namespaces=namespaces, **kwargs)

    async def get_radec(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current RA and Dec.

        Returns:
            Tuple of current RA and Dec in degrees.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError()

        # get RA/Dec
        return self._status['POSITION.EQUATORIAL.RA_J2000'] * 15., self._status['POSITION.EQUATORIAL.DEC_J2000']

    async def get_altaz(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current Alt and Az.

        Returns:
            Tuple of current Alt and Az in degrees.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError

        # get Alt/Az
        return self._status['POSITION.HORIZONTAL.ALT'], self._status['POSITION.HORIZONTAL.AZ']

    async def list_filters(self, **kwargs: Any) -> List[str]:
        """List available filters.

        Returns:
            List of available filters.
        """
        return self._filters

    async def get_filter(self, **kwargs: Any) -> str:
        """Get currently set filter.

        Returns:
            Name of currently set filter.
        """
        return await self._pilar.filter_name()

    @timeout(60000)
    async def set_filter(self, filter_name: str, **kwargs: Any) -> None:
        """Set the current filter.

        Args:
            filter_name: Name of filter to set.

        Raises:
            ValueError: If binning could not be set.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # acquire lock
        async with LockWithAbort(self._lock_filter, self._abort_filter):
            log.info('Changing filter to %s...', filter_name)
            await self._change_motion_status(MotionStatus.SLEWING, interface='IFilters')
            await self._pilar.change_filter(filter_name, force_forward=self._force_filter_forward,
                                            abort_event=self._abort_filter)
            await self._change_motion_status(MotionStatus.POSITIONED, interface='IFilters')
            log.info('Filter changed.')

            # send event
            await self.comm.send_event(FilterChangedEvent(current=filter_name))

    async def _move_altaz(self, alt: float, az: float, abort_event: asyncio.Event) -> None:
        """Actually moves to given coordinates. Must be implemented by derived classes.

        Args:
            alt: Alt in deg to move to.
            az: Az in deg to move to.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            Exception: On error.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # reset offsets
        await self._reset_offsets()

        # start tracking
        await self._change_motion_status(MotionStatus.SLEWING, interface='ITelescope')
        success = await self._pilar.goto(alt, az, abort_event=abort_event)
        await self._change_motion_status(MotionStatus.POSITIONED, interface='ITelescope')

        # finished
        if success:
            log.info('Reached destination.')
        else:
            raise ValueError('Could not reach destination.')

    async def _move_radec(self, ra: float, dec: float, abort_event: asyncio.Event) -> None:
        """Actually starts tracking on given coordinates. Must be implemented by derived classes.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            Exception: On any error.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # reset offsets
        await self._reset_offsets()

        # start tracking
        await self._change_motion_status(MotionStatus.SLEWING, interface='ITelescope')
        success = await self._pilar.track(ra, dec, abort_event=abort_event)
        await self._change_motion_status(MotionStatus.TRACKING, interface='ITelescope')

        # finished
        if success:
            log.info('Reached destination.')
        else:
            raise ValueError('Could not reach destination.')

    async def _reset_offsets(self) -> None:
        """Reset Alt/Az offsets."""
        await self._pilar.set('POSITION.INSTRUMENTAL.ZD.OFFSET', 0)
        await self._pilar.set('POSITION.INSTRUMENTAL.AZ.OFFSET', 0)

    async def get_focus(self, **kwargs: Any) -> float:
        """Return current focus.

        Returns:
            Current focus.
        """
        return float(await self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.CURRPOS'))

    async def get_focus_offset(self, **kwargs: Any) -> float:
        """Return current focus offset.

        Returns:
            Current focus offset.
        """
        return float(await self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.OFFSET'))

    @timeout(30000)
    async def set_focus(self, focus: float, **kwargs: Any) -> None:
        """Sets new focus.

        Args:
            focus: New focus value.

        Raises:
            InterruptedError: If focus was interrupted.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # acquire lock
        async with LockWithAbort(self._lock_focus, self._abort_focus):
            # start
            log.info('Setting focus to %.4f...', focus)
            await self._change_motion_status(MotionStatus.SLEWING, interface='IFocuser')
            #self._pilar.set('POSITION.INSTRUMENTAL.FOCUS.TARGETPOS', focus,
            #                timeout=30000, abort_event=self._abort_focus)

            # set focus
            await self._pilar.focus(focus)

            # finished
            await self._change_motion_status(MotionStatus.POSITIONED, interface='IFocuser')
            log.info('Reached new focus of %.4f.',
                     float(await self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.CURRPOS')))

    @timeout(30000)
    async def set_focus_offset(self, offset: float, **kwargs: Any) -> None:
        """Sets focus offset.

        Args:
            offset: New focus offset.

        Raises:
            InterruptedError: If focus was interrupted.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # acquire lock
        async with LockWithAbort(self._lock_focus, self._abort_focus):
            # set focus
            log.info('Setting focus offset to %.2f...', offset)
            await self._change_motion_status(MotionStatus.SLEWING, interface='IFocuser')
            await self._pilar.set('POSITION.INSTRUMENTAL.FOCUS.OFFSET', offset,
                                  timeout=10000)
            await self._change_motion_status(MotionStatus.POSITIONED, interface='IFocuser')
            log.info('Reached new focus offset of %.2f.',
                     float(await self._pilar.get('POSITION.INSTRUMENTAL.FOCUS.OFFSET')))

    @timeout(10000)
    async def set_offsets_altaz(self, dalt: float, daz: float, **kwargs: Any) -> None:
        """Move an Alt/Az offset.

        Args:
            dalt: Altitude offset in degrees.
            daz: Azimuth offset in degrees.

        Raises:
            ValueError: If offset could not be set.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # set offsets
        log.info('Moving offset of dAlt=%.3f", dAz=%.3f".', dalt * 3600., daz * 3600.)
        await self.comm.send_event(OffsetsAltAzEvent(alt=dalt, az=daz))
        old_status = await self.get_motion_status(interface='ITelescope')
        await self._change_motion_status(MotionStatus.SLEWING, interface='ITelescope')
        await self._pilar.set('POSITION.INSTRUMENTAL.ZD.OFFSET', -dalt)
        await self._pilar.set('POSITION.INSTRUMENTAL.AZ.OFFSET', daz)

        # just wait a second and finish
        await asyncio.sleep(5)
        await self._change_motion_status(old_status, interface='ITelescope')

    async def get_offsets_altaz(self, **kwargs: Any) -> Tuple[float, float]:
        """Get Alt/Az offset.

        Returns:
            Tuple with alt and az offsets.
        """

        # get current offsets and return then
        dalt = -float(await self._pilar.get('POSITION.INSTRUMENTAL.ZD.OFFSET'))
        daz = float(await self._pilar.get('POSITION.INSTRUMENTAL.AZ.OFFSET'))
        return dalt, daz

    @timeout(300000)
    async def init(self, **kwargs: Any) -> None:
        """Initialize device.

        Raises:
            ValueError: If device could not be initialized.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # init telescope
        log.info('Initializing telescope...')
        await self._change_motion_status(MotionStatus.INITIALIZING)
        if not await self._pilar.init():
            await self._change_motion_status(MotionStatus.ERROR)
            raise ValueError('Could not initialize telescope.')

        # init filter wheel
        log.info('Initializing filter wheel...')
        await self.set_filter(self._filters[-1])
        await self.set_filter('clear')

        # finished, send event
        await self._change_motion_status(MotionStatus.IDLE)

    @timeout(300000)
    async def park(self, **kwargs: Any) -> None:
        """Park device.

        Raises:
            ValueError: If device could not be parked.
        """

        # check error
        if self._pilar.has_error:
            raise ValueError('Telescope in error state.')

        # reset all offsets
        await self._reset_offsets()

        # park telescope
        log.info('Parking telescope...')
        await self._change_motion_status(MotionStatus.PARKING)
        if not await self._pilar.park():
            await self._change_motion_status(MotionStatus.ERROR)
            raise ValueError('Could not park telescope.')
        await self._change_motion_status(MotionStatus.PARKED)

    async def get_temperatures(self, **kwargs: Any) -> Dict[str, float]:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """

        # get all temperatures
        temps = {}
        for name, var in self._temperatures.items():
            temps[name] = self._status[var]

        # return it
        return temps

    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        """Stop the motion.

        Args:
            device: Name of device to stop, or None for all.
        """
        await self._pilar.stop()
        await self._change_motion_status(MotionStatus.IDLE)
        log.info('Stopped all motion.')

    async def is_ready(self, **kwargs: Any) -> bool:
        """Returns the device is "ready", whatever that means for the specific device.

        Returns:
            Whether device is ready
        """

        # check that motion is not in one of the states listed below
        return await self.get_motion_status() not in [MotionStatus.PARKED, MotionStatus.INITIALIZING,
                                                      MotionStatus.PARKING, MotionStatus.ERROR, MotionStatus.UNKNOWN]

    async def start_pointing_series(self, **kwargs: Any) -> str:
        """Start a new pointing series.

        Returns:
            A unique ID or filename, by which the series can be identified.
        """

        # no path given?
        if self._pointing_path is None:
            raise ValueError('No path for pointing given in config.')
        log.info('Starting new pointing series...')

        # clear list of measurements
        await self._pilar.safe_set('POINTING.MODEL.CLEAR', 1, msg='Could not clear list of measurements: ')

        # set filename
        dt = datetime.datetime.utcnow().strftime('%Y%m%d-%H%M%S')
        filename = os.path.join(self._pointing_path, f'pointing_{dt}.dat')
        await self._pilar.safe_set('POINTING.MODEL.FILE', filename, msg='Could not set filename: ')

        # check it
        log.info(f'Checking filename: {await self._pilar.get("POINTING.MODEL.FILE")}.')

        # no auto-save
        await self._pilar.safe_set('POINTING.MODEL.AUTO_SAVE', 0, msg='Could not disable auto-saving: ')

        # finished
        return filename

    async def stop_pointing_series(self, **kwargs: Any) -> None:
        """Stop a pointing series."""

        # save model
        log.info('Stopping pointing series...')
        await self._pilar.safe_set('POINTING.MODEL.SAVE', 1)

    async def add_pointing_measure(self, **kwargs: Any) -> None:
        """Add a new measurement to the pointing series."""

        # add point
        log.info('Adding point to pointing series...')
        await self._pilar.safe_set('POINTING.MODEL.ADD', str(self._pointing_id), msg='Could not add measurement: ')
        self._pointing_id += 1

        # get number of points
        log.info(f'Number of measurements: {await self._pilar.get("POINTING.MODEL.COUNT")}.')


__all__ = ['PilarTelescope']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyobs_pilar']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyobs-pilar',
    'version': '0.15.0a4',
    'description': 'pyobs model for Pilar TCS',
    'long_description': None,
    'author': 'Tim-Oliver Husser',
    'author_email': 'thusser@uni-goettingen.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
}


setup(**setup_kwargs)
